namespace Syntriax.Engine.Core;

/// <summary>
/// Represents an HSV color.
/// </summary>
/// <param name="hue">Hue of the <see cref="ColorHSV"/>.</param>
/// <param name="saturation">Saturation of the <see cref="ColorHSV"/>.</param>
/// <param name="value">Value of the <see cref="ColorHSV"/>.</param>
/// <remarks>
/// Initializes a new instance of the <see cref="ColorHSV"/> struct with the specified values.
/// </remarks>
[System.Diagnostics.DebuggerDisplay("{ToString(),nq}")]
public readonly struct ColorHSV(float hue, float saturation, float value)
{
    /// <summary>
    /// The Hue value of the <see cref="ColorHSV"/>.
    /// </summary>
    public readonly float Hue = hue.Clamp(0f, 1f);

    /// <summary>
    /// The Saturation value of the <see cref="ColorHSV"/>.
    /// </summary>
    public readonly float Saturation = saturation.Clamp(0f, 1f);

    /// <summary>
    /// The Value value of the <see cref="ColorHSV"/>.
    /// </summary>
    public readonly float Value = value.Clamp(0f, 1f);

    public static ColorHSV operator -(ColorHSV color) => new(color.Hue.OneMinus(), color.Saturation.OneMinus(), color.Value.OneMinus());
    public static ColorHSV operator +(ColorHSV left, ColorHSV right) => new(left.Hue + right.Hue, left.Saturation + right.Saturation, left.Value + right.Value);
    public static ColorHSV operator -(ColorHSV left, ColorHSV right) => new(left.Hue - right.Hue, left.Saturation - right.Saturation, left.Value - right.Value);
    public static ColorHSV operator *(ColorHSV left, ColorHSV right) => new(left.Hue * right.Hue, left.Saturation * right.Saturation, left.Value * right.Value);
    public static ColorHSV operator *(ColorHSV color, float value) => new(color.Hue * value, color.Saturation * value, color.Value * value);
    public static ColorHSV operator *(float value, ColorHSV color) => new(color.Hue * value, color.Saturation * value, color.Value * value);
    public static ColorHSV operator /(ColorHSV color, float value) => new(color.Hue / value, color.Saturation / value, color.Value / value);
    public static bool operator ==(ColorHSV left, ColorHSV right) => left.Hue == right.Hue && left.Saturation == right.Saturation && left.Value == right.Value;
    public static bool operator !=(ColorHSV left, ColorHSV right) => left.Hue != right.Hue || left.Saturation != right.Saturation || left.Value != right.Value;

    public static implicit operator ColorHSV(ColorHSVA hsva) => new(hsva.Hue, hsva.Saturation, hsva.Value);
    public static implicit operator ColorHSV(ColorRGBA rgba) => (ColorHSVA)rgba;
    public static implicit operator ColorHSV(ColorRGB rgb) => (ColorHSVA)rgb;

    /// <summary>
    /// Inverts the given <see cref="ColorHSV"/>.
    /// </summary>
    /// <param name="color">The <see cref="ColorHSV"/>.</param>
    /// <returns>The inverted <see cref="ColorHSV"/>.</returns>
    public static ColorHSV Invert(ColorHSV color) => -color;

    /// <summary>
    /// Adds two <see cref="ColorHSV"/>s.
    /// </summary>
    /// <param name="left">The first <see cref="ColorHSV"/>.</param>
    /// <param name="right">The second <see cref="ColorHSV"/>.</param>
    /// <returns>The sum of the two <see cref="ColorHSV"/>s.</returns>
    public static ColorHSV Add(ColorHSV left, ColorHSV right) => left + right;

    /// <summary>
    /// Subtracts one <see cref="ColorHSV"/> from another.
    /// </summary>
    /// <param name="left">The <see cref="ColorHSV"/> to subtract from.</param>
    /// <param name="right">The <see cref="ColorHSV"/> to subtract.</param>
    /// <returns>The result of subtracting the second <see cref="ColorHSV"/> from the first.</returns>
    public static ColorHSV Subtract(ColorHSV left, ColorHSV right) => left - right;

    /// <summary>
    /// Multiplies a <see cref="ColorHSV"/> by a scalar value.
    /// </summary>
    /// <param name="color">The <see cref="ColorHSV"/>.</param>
    /// <param name="value">The scalar value.</param>
    /// <returns>The result of multiplying the <see cref="ColorHSV"/> by the scalar value.</returns>
    public static ColorHSV Multiply(ColorHSV color, float value) => color * value;

    /// <summary>
    /// Divides a <see cref="ColorHSV"/> by a scalar value.
    /// </summary>
    /// <param name="color">The <see cref="ColorHSV"/>.</param>
    /// <param name="value">The scalar value.</param>
    /// <returns>The result of dividing the <see cref="ColorHSV"/> by the scalar value.</returns>
    public static ColorHSV Divide(ColorHSV color, float value) => color / value;

    /// <summary>
    /// Performs linear interpolation between two <see cref="ColorHSV"/>s.
    /// </summary>
    /// <param name="from">The starting <see cref="ColorHSV"/> (t = 0).</param>
    /// <param name="to">The ending <see cref="ColorHSV"/> (t = 1).</param>
    /// <param name="t">The interpolation parameter.</param>
    /// <returns>The interpolated <see cref="ColorHSV"/>.</returns>
    public static ColorHSV Lerp(ColorHSV from, ColorHSV to, float t)
    {
        float hueDiff = to.Hue - from.Hue;
        float saturationDiff = to.Saturation - from.Saturation;
        float valueDiff = to.Value - from.Value;

        return from + new ColorHSV(hueDiff * t, saturationDiff * t, valueDiff * t);
    }

    /// <summary>
    /// Checks if two <see cref="ColorHSV"/>s are approximately equal within a specified epsilon range.
    /// </summary>
    /// <param name="left">The first <see cref="ColorHSV"/>.</param>
    /// <param name="right">The second <see cref="ColorHSV"/>.</param>
    /// <param name="epsilon">The epsilon range.</param>
    /// <returns><see cref="true"/> if the <see cref="ColorHSV"/>s are approximately equal; otherwise, <see cref="false"/>.</returns>
    public static bool ApproximatelyEquals(ColorHSV left, ColorHSV right, float epsilon = float.Epsilon)
        => left.Hue.ApproximatelyEquals(right.Hue, epsilon) && left.Saturation.ApproximatelyEquals(right.Saturation, epsilon) && left.Value.ApproximatelyEquals(right.Value, epsilon);

    /// <summary>
    /// Determines whether the specified object is equal to the current <see cref="ColorHSV"/>.
    /// </summary>
    /// <param name="obj">The object to compare with the current <see cref="ColorHSV"/>.</param>
    /// <returns><see cref="true"/> if the specified object is equal to the current <see cref="ColorHSV"/>; otherwise, <see cref="false"/>.</returns>
    public override bool Equals(object? obj) => obj is ColorHSV colorHSV && this == colorHSV;

    /// <summary>
    /// Generates a hash code for the <see cref="ColorHSV"/>.
    /// </summary>
    /// <returns>A hash code for the <see cref="ColorHSV"/>.</returns>
    public override int GetHashCode() => System.HashCode.Combine(Hue, Saturation, Value);

    /// <summary>
    /// Converts the <see cref="ColorHSV"/> to its string representation.
    /// </summary>
    /// <returns>A string representation of the <see cref="ColorHSV"/>.</returns>
    public override string ToString() => $"{nameof(ColorHSV)}({Hue}, {Saturation}, {Value})";
}

/// <summary>
/// Provides extension methods for <see cref="ColorHSV"/> type.
/// </summary>
public static class ColorHSVExtensions
{
    /// <inheritdoc cref="ColorHSV.Add(ColorHSV, ColorHSV)" />
    public static ColorHSV Add(this ColorHSV color, ColorHSV value) => ColorHSV.Add(color, value);

    /// <inheritdoc cref="ColorHSV.Subtract(ColorHSV, ColorHSV)" />
    public static ColorHSV Subtract(this ColorHSV color, ColorHSV value) => ColorHSV.Subtract(color, value);

    /// <inheritdoc cref="ColorHSV.Multiply(ColorHSV, ColorHSV)" />
    public static ColorHSV Multiply(this ColorHSV color, float value) => ColorHSV.Multiply(color, value);

    /// <inheritdoc cref="ColorHSV.Divide(ColorHSV, ColorHSV)" />
    public static ColorHSV Divide(this ColorHSV color, float value) => ColorHSV.Divide(color, value);

    /// <inheritdoc cref="ColorHSV.Lerp(ColorHSV, ColorHSV, float)" />
    public static ColorHSV Lerp(this ColorHSV from, ColorHSV to, float t) => ColorHSV.Lerp(from, to, t);

    /// <inheritdoc cref="ColorHSV.ApproximatelyEquals(ColorHSV, ColorHSV, float) " />
    public static bool ApproximatelyEquals(this ColorHSV left, ColorHSV right, float epsilon = float.Epsilon) => ColorHSV.ApproximatelyEquals(left, right, epsilon);
}
